#!/usr/bin/env python3
import os
import errno
import json
import getpass
import argparse

from . import config
from . import track_cli


def main():
    parser = argparse.ArgumentParser(
        description='Track completion of your jobs!'
    )

    parser.add_argument(
        '-c',
        '--command',
        nargs='+',
        help='full command to track',
    )

    # May potentially support multiple processes
    parser.add_argument(
        '-p',
        '--pid',
        nargs=1,
        type=int,
        help='PID of existing job to track',
    )

    # Send SMS at job completion
    parser.add_argument('-s',
                        '--sms',
                        action='store_true',
                        help='send SMS at job exit')
    
    # Send email at job completion
    parser.add_argument('-e',
                        '--email',
                        action='store_true',
                        help='send email at job exit')
    
    # Init / info
    parser.add_argument('-i',
                        '--init',
                        action='store_true',
                        help='enter information needed for notifcations')

    # Verbose
    parser.add_argument('-v',
                        '--verbose',
                        action='store_true',
                        help='print payloads and all info')

    args = parser.parse_args()

    if args.init:
        with open(f"{config.JORT_DIR}/config", "r") as f:
            try:
                config_data = json.load(f)
            except json.decoder.JSONDecodeError:
                config_data = {}
        input_config_data = {
            "email": input('What email to use? ({}) '
                           .format(config_data.get("email"))),
            "twilio_receive_number": input('What phone number to receive SMS? ({}) '
                                           .format(config_data.get("twilio_receive_number"))),
            "twilio_send_number": input('What Twilio number to send SMS? ({}) '
                                        .format(config_data.get("twilio_send_number"))),
            "twilio_account_sid": input('Twilio Account SID? ({}) '
                                        .format(config_data.get("twilio_account_sid"))),
            "twilio_auth_token": getpass.getpass('Twilio Auth Token? ({}) '
                                                 .format("*"*len(config_data.get("twilio_auth_token", "")))),
        }
        for key in input_config_data:
            if input_config_data[key] != "":
                config_data[key] = input_config_data[key]
        with open(f"{config.JORT_DIR}/config", "w") as f:
            json.dump(config_data, f)
                
    if args.command and args.pid:
        parser.error('Please specify only one command or process to track.')
    elif args.command is None and args.pid is None and not args.init:
        parser.print_help()
    elif args.command:
        # # Grab all aws credentials; either from file or interactively
        # aws_credentials = auth.login()
        joined_command = ' '.join(args.command)
        print(f"Tracking command '{joined_command}'")
        track_cli.track_new(joined_command,
                            store_stdout=False,
                            save_filename=None,
                            send_sms=args.sms,
                            send_email=args.email,
                            verbose=args.verbose)
    elif args.pid:
        # # Grab all aws credentials; either from file or interactively
        # aws_credentials = auth.login()
        print(f"Tracking existing process PID at: {args.pid[0]}")
        track_cli.track_existing(args.pid[0],
                                 send_sms=args.sms,
                                 send_email=args.email,
                                 verbose=args.verbose)
    elif args.init:
        pass
    else:
        parser.error('Something went wrong!')

if __name__ == '__main__':
    main()


import sys
import time
import logging
from functools import wraps

from . import checkpoint
from . import datetime_utils
        

class Tracker(object):
    def __init__(self, logname="tracker.log", verbose=0):
        """
        Create tracker object. Arguments give logging options; verbosity 0 for none, 
        1 for INFO, and 2 for DEBUG.
        """
        self.date_created = datetime_utils.get_iso_date()
        self.checkpoints = {}
        self.open_checkpoint_payloads = {}
        self.logname = logname
        if verbose != 0:
            if verbose == 1:
                level = logging.INFO
            else:
                level = logging.DEBUG
            file_handler = logging.FileHandler(filename=self.logname, mode="w")
            stdout_handler = logging.StreamHandler(sys.stdout)
            handlers = [file_handler, stdout_handler]

            logging.basicConfig(level=level,
                                format="%(asctime)s %(name)-15s %(levelname)-8s %(message)s",
                                handlers=handlers, 
                                force=True)
        
    def start(self, name=None, date_created=None):
        if name == None:
            name = "Misc"
        if name in self.open_checkpoint_payloads:
            raise RuntimeError(f"Open checkpoint named {name} already exists")
        
        if date_created is not None:
            start = date_created
        else:
            start = datetime_utils.get_iso_date()
        now = datetime_utils.get_iso_date()

        self.open_checkpoint_payloads[name] = {
            "user_id": None,
            "job_id": None,
            "short_name": name,
            "long_name": name,
            "status": "running",
            "machine": None,
            "date_created": start,
            "date_modified": now,
            "runtime": datetime_utils.get_runtime(start, now),
            "stdout_fn": None,
            "unread": True,
            "error_message": None,
        }

        if name not in self.checkpoints:
            self.checkpoints[name] = checkpoint.Checkpoint(name)
        logger = logging.getLogger(f"{name}.start")
        logger.debug("Profiling block started.")
        
    def stop(self, name=None, callbacks=[]):
        if name == None:
            name = list(self.open_checkpoint_payloads.keys())[-1]
        elif name not in self.open_checkpoint_payloads:
            raise KeyError(f"No open checkpoint named {name}")

        payload = self.open_checkpoint_payloads.pop(name)
        if payload["status"] == "running":
            payload["status"] = "success"
        start = payload["date_created"]
        stop = datetime_utils.update_payload_times(payload)
        self.checkpoints[name].add_times(start, stop)

        logger = logging.getLogger(f"{name}.stop")
        logger.debug("Profiling block stopped.")
        formatted_runtime = checkpoint.format_reported_times(self.checkpoints[name].elapsed[-1])
        logger.info(f"Elapsed time: {formatted_runtime}")

        for callback in callbacks:
            callback.execute(payload=payload)
        
    def remove(self, name=None):
        """
        Option to remove checkpoint start instead of completing a profiling
        set, for example on catching an error.
        """
        if name == None:
            name = list(self.open_checkpoint_payloads.keys())[-1]
        
        if name in self.open_checkpoint_payloads:
            payload = self.open_checkpoint_payloads.pop(name)
            logger = logging.getLogger(f"{name}.remove")
            logger.debug("Profiling block removed.")
        
    def clear_open(self):
        self.open_checkpoint_payloads = {}
        
    def time_func(self, f, report=False):
        """
        Function wrapper for tracker.
        """
        @wraps(f)
        def wrapper(*args, **kwargs):
            self.start(name=f.__qualname__)
            result = f(*args, **kwargs)
            self.stop(name=f.__qualname__)
            if report:
                self.report()
            return result
        return wrapper
        
    def report(self, dec=1):
        for name in self.checkpoints:
            ckpt = self.checkpoints[name]
            print(ckpt.report(dec=dec))
            
            
def time_func(f):
    """
    Independent function wrapper. Creates a one-off tracker and reports time.
    """
    return Tracker(verbose=0).time_func(f, report=True)
    # @wraps(f)
    # def wrapper(*args, **kwargs):
    #     tr.start(name=f.__qualname__)
    #     result = f(*args, **kwargs)
    #     tr.stop(name=f.__qualname__)
    #     tr.report()
    #     return result
    # return wrapper


from datetime import datetime
import dateutil.parser


def get_iso_date(timestamp=None):
    """Return start date in ISO 8601 format"""
    if timestamp:
        return datetime.utcfromtimestamp(timestamp).isoformat()
    else:
        return datetime.utcnow().isoformat()

def get_runtime(iso_date1, iso_date2):
    """
    Return runtime in seconds between two dates in ISO format.
    """
    runtime = (
        dateutil.parser.parse(iso_date2) - dateutil.parser.parse(iso_date1)
    )
    return runtime.total_seconds()

# def get_current_times(start_date):
#     """Return current date in ISO 8601 format, as well as runtime calculated
#     from the starting date"""
#     # Start time in ISO 8601
#     now_date = get_iso_date()
#     runtime_s = get_runtime(now_date, start_date)
#     hours, remainder = divmod(runtime_s, 3600)
#     minutes, seconds = divmod(remainder, 60)
#     formatted_runtime = f"{int(hours):02d}:{int(minutes):02d}:{int(seconds):02d}"

#     return now_date, formatted_runtime, runtime_s


def update_payload_times(payload):
    """
    Modify payload dictionary with current times. Returns current date.
    """
    date_now = get_iso_date()
    runtime = get_runtime(payload["date_created"], date_now)
    payload['runtime'] = runtime
    payload['date_modified'] = date_now
    return date_now


import os
import json
import twilio.rest
import humanfriendly
from . import config


class Callback(object):
    def __init__(self):
        pass

    def format_message(self, payload):
        pass

    def execute(self, payload):
        pass
    
    
class PrintReport(Callback):
    def __init__(self):
        pass

    def format_message(self, payload):
        return (
            f'The job \'{payload["short_name"]}\' finished running '
            f'in {humanfriendly.format_timespan(payload["runtime"])}'
        )

    def execute(self, payload):
        print(self.format_message(payload))


class SMSNotification(Callback):
    def __init__(self, receive_number=None):
        self.receive_number = receive_number
        if receive_number is None:
            with open(f"{config.JORT_DIR}/config", "r") as f:
                config_data = json.load(f)
                self.receive_number = config_data["twilio_receive_number"]
                self.send_number = config_data["twilio_send_number"]
                self.twilio_account_sid = config_data["twilio_account_sid"]
                self.twilio_auth_token = config_data["twilio_auth_token"]

    def format_message(self, payload):
        if payload["status"] == "success":
            return (
                f'Your job \'{payload["short_name"]}\' successfully completed '
                f'in {humanfriendly.format_timespan(payload["runtime"])}'
            )
        elif payload["status"] == "error":
            error_name = payload["error_message"].split(":")[0]
            return (
                f'Your job \'{payload["short_name"]}\' exited in error ({error_name}) '
                f'after {humanfriendly.format_timespan(payload["runtime"])}'
            )
        else:
            return (
                f'Your job \'{payload["short_name"]}\' finished running '
                f'in {humanfriendly.format_timespan(payload["runtime"])}'
            )
        
    
    def execute(self, payload):
        client = twilio.rest.Client(self.twilio_account_sid,
                                    self.twilio_auth_token)
        message = client.messages.create(body=self.format_message(payload),
                                         from_=self.send_number,
                                         to=self.receive_number)


class EmailNotification(Callback):
    def __init__(self, email=None):
        self.email = email
        if email is None:
            with open(f"{config.JORT_DIR}/config", "r") as f:
                config_data = json.load(f)
                self.email = config_data["email"]

    def format_message(self, payload):
        return ""

    def execute(self, payload):
        pass





import numpy as np
from . import datetime_utils


def format_reported_times(m, s=None, dec=1):
    format_str = ""
    if m >= 1.:
        format_str += f"{m:.{dec}f} s" 
        if s is not None:
            format_str += f" \u00B1 {s:.{dec}f} s"
    elif m >= 1e-3:
        format_str += f"{1e3 * m:.{dec}f} ms" 
        if s is not None:
            format_str += f" \u00B1 {1e3 * s:.{dec}f} ms"
    elif m >= 1e-6:
        format_str += f"{1e6 * m:.{dec}f} us" 
        if s is not None:
            format_str += f" \u00B1 {1e6 * s:.{dec}f} us"
    else:
        format_str += f"{1e9 * m:.{dec}f} ns" 
        if s is not None:
            format_str += f" \u00B1 {1e9 * s:.{dec}f} ns"
    return format_str


class Checkpoint(object):
    def __init__(self, name):
        """
        Create checkpoint to save times.

        Start and stop times are in ISO format, while elapsed times are in seconds.
        """
        self.name = name
        self.starts = []
        self.stops = []
        self.elapsed = []
        
    def _get_elapsed(self):
        self.elapsed = [datetime_utils.get_runtime(t1, t2) 
                        for (t1, t2) in zip(self.starts, self.stops)]
        
    def add_times(self, start, stop):
        self.starts.append(start)
        self.stops.append(stop)
        self.elapsed.append(datetime_utils.get_runtime(start, stop))
    
    def report(self, dec=1):
        """
        Return formatted statistics.
        """
        elapsed_stats = format_reported_times(np.mean(self.elapsed), 
                                              np.std(self.elapsed), 
                                              dec=dec)
        return (
            f"{self.name} | "
            f"{elapsed_stats} per iteration, "
            f"n = {len(self.elapsed)}"
        )
        
    def __str__(self):
        return self.report(dec=1)
    


from __future__ import absolute_import, division, print_function

from .tracker import Tracker, time_func
from .track_cli import track_new, track_existing
from .reporting_callbacks import EmailNotification, SMSNotification, PrintReport



import os
from pathlib import Path

# Create internal jort directory
JORT_DIR = f"{os.path.expanduser('~')}/.jort"
Path(f"{JORT_DIR}/").mkdir(mode=0o700, parents=True, exist_ok=True)
Path(f"{JORT_DIR}/config").touch(mode=0o600, exist_ok=True)


import os
import sys
import time
import subprocess
import uuid
import psutil
from tqdm import tqdm
from pprint import pprint

from . import tracker
from . import datetime_utils
from . import config
from . import reporting_callbacks


def track_new(command,
              store_stdout=False,
              save_filename=None,
              send_sms=False,
              send_email=False,
              verbose=False,
              update_period=-1):
    """
    Track execution time and details of new command line process.

    If update_period is negative, doesn't do payload updates.
    """
    callbacks = [reporting_callbacks.PrintReport()]
    if send_sms:
        callbacks.append(reporting_callbacks.SMSNotification())
    if send_email:
        callbacks.append(reporting_callbacks.EmailNotification())

    # Key for storing stdout text to file
    job_id = str(uuid.uuid4())
    if save_filename or store_stdout:
        stdout_fn = f"{job_id}.txt"
        stdout_path = f"{config.JORT_DIR}/{stdout_fn}"
    else:
        stdout_fn = None

    tr = tracker.Tracker()
    tr.start(name=command)

    payload = tr.open_checkpoint_payloads[command]

    payload['job_id'] = job_id
    payload['stdout_fn'] = stdout_fn

    # ACTUALLY START SUBPROCESS
    my_env = os.environ.copy()
    my_env["PYTHONUNBUFFERED"] = "1"

    p = psutil.Popen(command.split(),
                     env=my_env,
                     stdout=subprocess.PIPE,
                     stderr=subprocess.STDOUT,
                     bufsize=1,
                     universal_newlines=True)
    print(f"Subprocess PID: {p.pid}\n")

    # Create stdout file
    if verbose:
        pprint(payload)
    if save_filename or store_stdout:
        f = open(stdout_path, "a+")
        f.write(f"$ {command}\n")
        f.close()

    buffer = ""
    temp_start = time.time()
    for line in p.stdout:
        if update_period > 0 and time.time() - temp_start >= update_period:
            if verbose:
                print("Buffered! (Not sent)", [buffer])
            if save_filename or store_stdout:
                f = open(stdout_path, "a+")
                f.write(buffer)
                f.close()

            payload['status'] = 'running'
            datetime_utils.update_payload_times(payload)
            if verbose:
                pprint(payload)

            buffer = ""
            temp_start = time.time()

        sys.stdout.write(line)
        buffer += line

    if verbose:
        print("Buffered!", [buffer])
    if save_filename or store_stdout:
        f = open(stdout_path, "a+")
        f.write(buffer)
        f.close()

    if verbose:
        print("Exit code:", p.poll())

    if p.returncode in [0, None]:
        payload["status"] = "success"
    else:
        payload["status"] = "error"
        payload["error_message"] = line
    tr.stop(callbacks=callbacks)

    # print("")
    # if payload["runtime"] < 10:
    #     sys.exit("Job exited in 10 seconds -- no need to track!")

    # if store_stdout:
    #     s3.meta.client.upload_file(
    #         stdout_path,
    #         aws_credentials["bucket_name"],
    #         "private/%s/%s" % (aws_credentials["identity_id"], stdout_fn)
    #     )
    if verbose:
        pprint(payload)

    if save_filename or store_stdout:
        if save_filename:
            subprocess.call(["cp", stdout_path, save_filename])
        try:
            subprocess.call(["rm", stdout_path])
        except Exception as e:
            raise e


def track_existing(pid,
                   send_sms=False,
                   send_email=False,
                   verbose=False,
                   update_period=-1):
    """
    Track execution time and details of existing command line process.

    If update_period is negative, doesn't do payload updates.
    """
    callbacks = [reporting_callbacks.PrintReport()]
    if send_sms:
        callbacks.append(reporting_callbacks.SMSNotification())
    if send_email:
        callbacks.append(reporting_callbacks.EmailNotification())

    # Does not support stdout tracking
    stdout_fn = None

    # Create process based on PID and grab relevant information
    p = psutil.Process(pid)
    job_id = str(uuid.uuid4())
    command = " ".join(p.cmdline())

    tr = tracker.Tracker()
    tr.start(name=command, date_created=datetime_utils.get_iso_date(p.create_time()))
    payload = tr.open_checkpoint_payloads[command]

    if verbose:
        pprint(payload)

    temp_start = time.time()

    while p.is_running():
        if update_period > 0 and time.time() - temp_start >= update_period:
            payload["status"] = "running"
            datetime_utils.update_payload_times(payload)
            if verbose:
                pprint(payload)

            temp_start = time.time()

    payload["status"] = "finished"
    tr.stop(callbacks=callbacks)

    # print("")
    # if runtime_s < 60:
    #     sys.exit("Job exited in less than a minute -- no need to track!")

    if verbose:
        pprint(payload)


import setuptools

__version__ = '1.0.0'

with open('README.md', 'r') as fh:
    long_description = fh.read()

entry_points = {
    'console_scripts': [
        'jort = jort.jort_exe:main'
    ]
}

with open("requirements.txt", "r") as f:
    install_requires = f.readlines()

setuptools.setup(
    name='jort',
    version=__version__,
    author='Bryan Brzycki',
    author_email='bbrzycki@berkeley.edu',
    description='Script profiler with checkpoints',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/bbrzycki/jort',
    project_urls={
        'Source': 'https://github.com/bbrzycki/jort'
    },
    entry_points=entry_points,
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    classifiers=(
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ),
)
